import java.util.*;

public class AwareRobot
        extends Robot {
        
    private int m_startMoney;
    private int m_money;
    private Board m_board;
    private Simulator m_sim;
    private Genome m_genome;
    private int[] m_baseX;
    private int[] m_baseY;
    private HashMap m_robots=new HashMap();
    private PackageMemory m_packageMemory=new PackageMemory(-1, -1); // loc doesn't matter
    private Base m_bases[];
        
    public AwareRobot(int id, int x, int y, int capacity, int money, Genome g, Board board, Simulator sim) {
        super(id, x, y, capacity, new int[] {0});
        m_startMoney=money;
        m_money=money;
        m_genome=g;
        m_board=board;
        m_sim=sim;
        int baseLocations[]=m_board.getBaseLocations();
        int baseCount=baseLocations.length/2;
        m_bases=new Base[baseCount];
        int zz=0;
        for (int i=0; i<baseLocations.length; i+=2) {
            m_bases[zz++]=new Base(baseLocations[i], baseLocations[i+1]);
        }
    }
    
    public Base getBase(int x, int y) {
        for (int i=0; i<m_bases.length; i++) {
            if ((m_bases[i].getX()==x) && (m_bases[i].getY()==y)) {
                return m_bases[i];
            }
        }
        return null;
    }
    
    public Base getBaseWith(int packageId) {
        for (int i=0; i<m_bases.length; i++) {
            if (m_bases[i].hasPackage(packageId)) 
                return m_bases[i];
        }
        return null;
    }
    
    public Base[] getBases() {
        return m_bases;
    }
    
    public int getStartMoney() {
        return m_startMoney;
    }
    
    public int getMoney() {
        return m_money;
    }
    
    public void tryPick(int bid, int[] packageIds) {
        m_money-=bid;
        m_sim.requestPick(this, bid, packageIds);
    }
    
    public void tryDrop(int bid, int[] packageIds) {
        m_money-=bid;
        m_sim.requestDrop(this, bid, packageIds);
    }
    
    public void tryMove(int bid, int dir) {
        m_money-=bid;
        m_sim.requestMove(this, bid, dir);
    }

    public String getStats() {
        return super.getStats() + ",$=" + m_money + "/" + m_startMoney;
    }

    public Robot getRobot(int id) {
        if (id==getId()) return (Robot) this;
        return (Robot) m_robots.get(new Integer(id));
    }
    
    public Robot addRobot(int id, int x, int y) {
System.out.println("Adding robot " + id);
        Robot r=new Robot(id, x, y, getCapacity(), new int[] {0}); 
        m_robots.put(new Integer(id), r);
        return r;
    }
    
    public int[] getRobotIds() {
        int[] ids=new int[m_robots.size()];
        Iterator ri=m_robots.keySet().iterator();
        int i=0;
        while (ri.hasNext()) {
            ids[i++]=((Integer) ri.next()).intValue();
        }
        return ids;
    }
    
    public void removeRobot(int id) {
System.out.println("Removing robot " + id);
        m_robots.remove(new Integer(id));
    }
    
    

    // a                   b         c          d
    // ---------------------------------------------------------------
    // NEW_ROBOT_POSITION  robotId   xPosition  yPosition
    // ROBOT_ACTION        robotId   PICK       packageId
    // ROBOT_ACTION        robotId   DROP       packageId
    // ROBOT_ACTION        robotId   MOVE       NORTH|SOUTH|EAST|WEST
    // ROBOT_ACTION        robotId   NONE
    // packageId           weight    destX      destY
    //
    // Note: 
    //        - a is a packageId if it's non-negative.
    //        - constants are from Simulator
    public void respond(int[] a, int[] b, int[] c, int[] d) {
        // update my view of the world
        for (int i=0; i<a.length; i++) {
            if (a[i]==Simulator.NEW_ROBOT_POSITION) {
                if (b[i]==getId()) {
                    setX(c[i]);
                    setY(d[i]);
                } else {
                    Robot r=getRobot(b[i]);
                    if (r==null) {
                        r=addRobot(b[i],c[i],d[i]);
                    } else {
                        r.setX(c[i]);
                        r.setY(d[i]);
                    }
                }
            } else if (a[i]==Simulator.ROBOT_ACTION) {
                if (c[i]==Simulator.PICK) {
                    // if a base had this, get the info from the base memory..
                    // this should always be the case if it's me.. not
                    // necessarily if it's another robot
                    Base baseWith=getBaseWith(d[i]);
                    int destX, destY, weight;
                    if (baseWith==null) {
                        destX=-1; destY=-1; weight=-1;
                    } else {
                        destX=baseWith.getPackageDestX(d[i]);
                        destY=baseWith.getPackageDestY(d[i]);
                        weight=baseWith.getPackageWeight(d[i]);
                    }
                    getRobot(b[i]).didPick(d[i], destX, destY, weight);
                    // if this was from a base location, remove the package
                    // from the memory for that base.
                    if (baseWith!=null) {
                        baseWith.removePackage(d[i]);
                    }
                } else if (c[i]==Simulator.DROP) {
                    getRobot(b[i]).didDrop(d[i]);
                } else if (c[i]==Simulator.MOVE) {
                    getRobot(b[i]).didMove(d[i]);
                }
            } else {
                // must be package info... we know this comes last, so i've
                // already moved.  if we're not at a base, this gets ignored
                // but if we're at one (getBase(x,y)), make sure it's added
                // to memory for that base.  (note: removal from bases
                // should occur through robot's various pick commands)
                Base thisBase=getBase(getX(), getY());
                if (thisBase!=null) {
                    thisBase.putPackage(a[i],b[i],c[i],d[i]);
                }
            }
        }
        // see if we're at a base, if so say we've visited it
                Base thisBase=getBase(getX(), getY());
                if (thisBase!=null) {
                    thisBase.visit();
                }
        // remove robots no longer on the board
        int[] rids=getRobotIds();
        for (int rIndex=0; rIndex<rids.length; rIndex++) {
            boolean found=false;
            for (int i=0; i<a.length; i++) {
                if (a[i]<0) { // if it's a robot thing
                    if (b[i]==rids[rIndex]) found=true;
                }
            }
            if (!found)
                removeRobot(rids[rIndex]);
        }

        // here's where i decide what to do.
        /*
        if (onDestination()) {
            dropPackages();
        } else {
            if (atBase() && baseHasPackagesIHaveRoomFor() ) {
                    pickBestPackages();
            } else {
                if (havePackages()) {
                    moveTowardBestDelivery();
                } else {
                    moveTowardBestBase();
                }
            }
        }
        */
        
        if (!dropIfAppropriate()) {
            if (!pickIfAppropriate()) {
                makeBestMove();
            }
        }
        
    }

    // we are not planning on picking or dropping, so decide what direction to head
    public void makeBestMove() {
        int[] pids=getPackageIds();
        int leastCostPid=-1;
        GridGraph g=m_board.getGridGraph();
        int dropPid=-1;
        int[] bestPath=null;
        // should we deliver?
        for (int i=0; i<pids.length; i++) {
            int[] p=g.getBestPath(getX()-1, getY()-1, getPackageDestX(pids[i])-1, getPackageDestY(pids[i])-1);
            if (p==null) {
                dropPid=pids[i];
            } else {
                int thisCost=g.getCost(getPackageDestX(pids[i])-1, getPackageDestY(pids[i])-1, getX()-1, getY()-1);
                if (leastCostPid==-1 && thisCost>0) {
                    leastCostPid=thisCost; bestPath=p;
                } else if (leastCostPid>thisCost) {
                    leastCostPid=thisCost; bestPath=p;
                }
            }
        }
        if (dropPid!=-1) {
            // drop cuz undeliverable -- don't know why
            tryDrop(1, new int[] {dropPid});
        } else if (leastCostPid==-1) {
            // none to deliver, go hunting!
            // hardcoded strategy:
            //  head toward closest base not known to be empty
            //  else chase after the last robot with the most known weight
            //  if all else fails, head north.
            Base closestUnvisitedOrNonEmptyBase=null;
            int bestCost=-1;
            for (int i=0; i<m_bases.length; i++) {
                if ((m_bases[i].getKnownPackageWeight()>0)||(!m_bases[i].visited())) {
                    int[] p=g.getBestPath(getX()-1, getY()-1, m_bases[i].getX()-1, m_bases[i].getY()-1);
                    if (p!=null) {
                        int thisCost=g.getCost(m_bases[i].getX()-1, m_bases[i].getY()-1, getX()-1, getY()-1);
                        if (bestCost==-1 && thisCost>0) {
                            bestCost=thisCost; bestPath=p;
                        } else if (bestCost>thisCost) {
                            bestCost=thisCost; bestPath=p;
                        }
                    }
                }
            }
            if (bestCost==-1) {
                // get most known weight bot
                int heaviestWeight=-1;
                int rids[]=getRobotIds();
                for (int i=0; i<rids.length; i++) {
                    int wt=getRobot(rids[i]).getKnownPackageWeight();
                    if (wt>heaviestWeight) {
                        // if i can reach it, assign it
                        int[] p=g.getBestPath(getX()-1, getY()-1, getRobot(rids[i]).getX()-1, getRobot(rids[i]).getY()-1);
                        if (p!=null) {
                            heaviestWeight=wt; bestPath=p;
                        }
                    }
                }
                if (heaviestWeight!=-1) {
                    // found one to annoy..head that way
                    System.out.println("((Found another robot to annoy, cuz couldn't find a non-empty base to get packages))");
                    if (bestPath[0]+1!=getX()) {
                        if (bestPath[0]+1>getX()) {
                            tryMove(1, Simulator.EAST);
                        } else {
                            tryMove(1, Simulator.WEST);
                        }
                    } else {
                        if (bestPath[1]+1>getY()) {
                            tryMove(1, Simulator.NORTH);
                        } else {
                            tryMove(1, Simulator.SOUTH);
                        }
                    }
                } else {
                    // didn't find one... act like an idiot and just walk north
                    System.out.println("((CONFUSED:ACTING LIKE AN IDIOT AND WALKING NORTH))");
                    tryMove(1, Simulator.NORTH);
                }
            } else {
                    System.out.println("((Heading toward a base that's not known to be empty))");
                if (bestPath[0]+1!=getX()) {
                    if (bestPath[0]+1>getX()) {
                        tryMove(1, Simulator.EAST);
                    } else {
                        tryMove(1, Simulator.WEST);
                    }
                } else {
                    if (bestPath[1]+1>getY()) {
                        tryMove(1, Simulator.NORTH);
                    } else {
                        tryMove(1, Simulator.SOUTH);
                    }
                }
            }
        } else {
            // head toward leastCostPid package's destination location -- hoping to deliver it!
                    System.out.println("((Heading toward one of my package's delivery destinations))");
            if (bestPath[0]+1!=getX()) {
                if (bestPath[0]+1>getX()) {
                    tryMove(1, Simulator.EAST);
                } else {
                    tryMove(1, Simulator.WEST);
                }
            } else {
                if (bestPath[1]+1>getY()) {
                    tryMove(1, Simulator.NORTH);
                } else {
                    tryMove(1, Simulator.SOUTH);
                }
            }
        }
    }
    
    public boolean pickIfAppropriate() {
        Base thisBase=getBase(getX(), getY());
        if (thisBase!=null) {
            int canCarry=getCapacity()-getKnownPackageWeight();
            // pick simply.. no time for bestFit or travel cost computation
            ArrayList carryList=new ArrayList();
            int[] thisBasePackageIds=thisBase.getPackageIds();
            for (int i=0; i<thisBasePackageIds.length; i++) {
                int weight=thisBase.getPackageWeight(thisBasePackageIds[i]);
                if (weight<canCarry) {
                    canCarry=canCarry-weight;
                    carryList.add(new Integer(thisBasePackageIds[i]));
                }
            }
            if (carryList.size()>0) {
                Iterator idIter=carryList.iterator();
                int[] pickIds=new int[carryList.size()];
                int i=0;
                while (idIter.hasNext()) {
                    pickIds[i++]=((Integer)idIter.next()).intValue();
                }
                    System.out.println("((Picking up one or more packages at a base))");
                tryPick(1, pickIds);
                return true;
            } else {
                return false;  // none fit
            }
        } else {
            return false;  // don't pick if not at base
        }
    }
    
    public boolean dropIfAppropriate() {
        // find out if i've got any packages to be delivered to this space,
        // and drop them if it seems like a good idea.
        int[] ids=getPackageIds();
        ArrayList droppable=new ArrayList();
        for (int i=0; i<ids.length; i++) {
            int destX=getPackageDestX(ids[i]);
            int destY=getPackageDestY(ids[i]);
            if (destX==getX() && destY==getY()) {
                droppable.add(new Integer(ids[i]));
            }
        }
        if (droppable.size()>0) {
            int[] dropIds=new int[droppable.size()];
            Iterator iter=droppable.iterator();
            int i=0;
            while (iter.hasNext()) {
                dropIds[i++]=((Integer) iter.next()).intValue();
            }
                    System.out.println("((Dropping one ore more packages cuz I think this is where it/they goes/go))");
            tryDrop(1, dropIds);
            return true;
        } else {
            return false;
        }
    }
    
}
