import java.io.BufferedReader;
import java.io.DataOutputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.net.Socket;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.StringTokenizer;

public class ICFPSimulatorProxy 
        extends Simulator {

    private HashSet m_players;
    private HashMap m_dataOutputStreams;
    private HashMap m_bufferedReaders;
    private HashMap m_nextRequestedAction;
    private String m_host;
    private int m_port;
    
    public ICFPSimulatorProxy(String host, int port) {
        m_host=host;
        m_port=port;
        m_players=new HashSet();
        m_dataOutputStreams=new HashMap();
        m_bufferedReaders=new HashMap();
        m_nextRequestedAction=new HashMap();
    }

    public void requestPick(AwareRobot robot, int bid, int packageIds[]) {
        StringBuffer b=new StringBuffer();
        b.append(bid);
        b.append(" Pick");
        for (int i=0; i<packageIds.length; i++) {
            b.append(' ');
            b.append(packageIds[i]);
        }
        System.out.println("----------->" + b.toString());
        m_nextRequestedAction.put(robot, b.toString());
    }
    
    public void requestDrop(AwareRobot robot, int bid, int packageIds[]) {
        StringBuffer b=new StringBuffer();
        b.append(bid);
        b.append(" Drop");
        for (int i=0; i<packageIds.length; i++) {
            b.append(' ');
            b.append(packageIds[i]);
        }
        System.out.println("----------->" + b.toString());
        m_nextRequestedAction.put(robot, b.toString());
    }
    
    public void requestMove(AwareRobot robot, int bid, int direction) {
        StringBuffer b=new StringBuffer();
        b.append(bid);
        b.append(" Move ");
        if (direction==NORTH) {
            b.append('N');
        } else if (direction==SOUTH) {
            b.append('S');
        } else if (direction==EAST) {
            b.append('E');
        } else {
            b.append('W');
        }
        System.out.println("----------->" + b.toString());
        m_nextRequestedAction.put(robot, b.toString());
    }
    
    public AwareRobot addPlayer(Genome g) {
      Socket s=null;
      DataOutputStream o=null;
      BufferedReader i=null;
      try {
        
        s=new Socket(m_host, m_port); 
        o=new DataOutputStream(s.getOutputStream()); 
        i=new BufferedReader(new InputStreamReader(s.getInputStream())); 
        o.writeBytes("Player\n"); 
        String dims=i.readLine();
        if (dims==null) {
            throw new IOException("Got null response from server after sending 'Player' line.");
        }
        int spacePos=dims.indexOf(" ");
        if (spacePos==-1) {
            throw new IOException("Got bad board dims '" + dims + "' response from server after sending 'Player' line.");
        }
        int xSize=0;
        int ySize=0;
        try {
            xSize=Integer.parseInt(dims.substring(0, spacePos));
            ySize=Integer.parseInt(dims.substring(spacePos+1));
        } catch (NumberFormatException nfe) {
            throw new IOException("Got bad board dims '" + dims + "' response (not number(s)) from server after sending 'Player' line.");
        }
        String[] boardInput=new String[ySize];
        for (int y=0; y<ySize; y++) {
            boardInput[y]=i.readLine();
            if (boardInput[y]==null) {
                throw new IOException("Got less than expected (" + ySize + ") lines of board from server, then it sent null.");
            }
        }
        String configLine=i.readLine();
        if (configLine==null) {
            throw new IOException("Got null response from server after reading board.");
        }
        spacePos=configLine.indexOf(" ");
        int lastSpacePos=configLine.lastIndexOf(" ");
        if ((spacePos==-1)||(lastSpacePos==-1)||(lastSpacePos<=spacePos)) {
            throw new IOException("Got bad player config '" + configLine + "' response from server after reading board.");
        }
        int id=0;
        int capacity=0;
        int money=0;
        try {
            id=Integer.parseInt(configLine.substring(0,spacePos));
            capacity=Integer.parseInt(configLine.substring(spacePos+1, lastSpacePos));
            money=Integer.parseInt(configLine.substring(lastSpacePos+1));
        } catch (NumberFormatException nfe) {
            throw new IOException("Got bad player config '" + configLine + "' response (not number(s)) from server after reading board.");
        }
        
        AwareRobot robot=new AwareRobot(id, 0, 0, capacity, money, g, 
                new Board(boardInput), (Simulator) this);
        m_players.add(robot);
        m_dataOutputStreams.put(robot, o);
        m_bufferedReaders.put(robot, i);
        sendServerReply(i, robot); // <-- for isaac
        return robot;
      } catch (IOException ioe) {
          System.err.println(ioe.getMessage());
          if (s!=null) {
              try {
              s.close();
              } catch (IOException ioe2) { }
          }
          return null;
      }
    }
    
    private void sendServerReply(BufferedReader in, AwareRobot robot) 
            throws IOException {
        // get the two response lines from the server and send them to
        // robot.
        //
        // The lines are:
        // 1) actions
        // 2) packages (if in player's current position)
        String actionLine=in.readLine();
        String packageLine=in.readLine();
        if (actionLine==null || packageLine==null) {
            throw new IOException("Could not get server response for robot " 
                    + robot.getId() + ", (was null) socket probably closed.");
        }
        if (actionLine.indexOf('#')!=0) {
            throw new IOException("Could not get server response for robot " 
                    + robot.getId() + ", (no # char) socket probably closed.");
        }
        System.out.println("Action line: '" + actionLine + "'");
        // first scan em both to see how much space we need for the arrays
        int numThings=numOccurs('N', actionLine) + numOccurs('S', actionLine)
                + numOccurs('E', actionLine) + numOccurs('W', actionLine)
                + numOccurs('X', actionLine) + numOccurs('P', actionLine)
                + numOccurs('D', actionLine) + numNonActions(actionLine);
                
        
        
//        System.out.println(numThings + " things happened.");
        
        int numPackages=(numOccurs(' ', packageLine) + 1) / 4;
        
//        System.out.println(numPackages + " packages are here.");
        
        
        int[] a= new int[numThings+numPackages]; 
        int[] b= new int[numThings+numPackages]; 
        int[] c= new int[numThings+numPackages]; 
        int[] d= new int[numThings+numPackages]; 

        // first parse the action line
        System.out.println("Parsing actionLine '" + actionLine + "'");
        int itemNum;
        int charNum=0;
        
        /*
        scanning for three types of things:
          "#(num)"              // set current robotnum
          
            "N|S|E|W"            // after looking at any of these,
            "X (num) Y (num)"
            "D|P (num)"
            
          " "#|N|S|E|W|X|D       // skip the space, 
          
        */
        int currentRobotId=-1;
        for (itemNum=0; itemNum<numThings; itemNum++) {
            char thisChar=actionLine.charAt(charNum);
//            System.out.println("itemIndex=" + itemNum + " numThings=" + numThings + " char='" + thisChar + "'");
            if (thisChar==' ') {
                thisChar=actionLine.charAt(++charNum);
            }
            if (thisChar=='#') {
// System.out.println("got to #");
                // scan till non-num (or end of string)
                // if itemNum==numThings-1,
                //   put the none action in 
                // else
                //   if charAt(charNum+1)=='#'
                //     it's a none action, add none action, then charNum++
                //   else
                //      
                int[] nn=scanTillNotInt(actionLine, charNum+1);
                currentRobotId=nn[0];
                if ((itemNum+1) == numThings) {
// System.out.println("got to ... must be looking at last thing");
                    if (nn[1]>actionLine.length()-1) {
// System.out.println("got to ... looks like a nonaction 1");
                        a[itemNum]=ROBOT_ACTION;
                        b[itemNum]=currentRobotId;
                        c[itemNum]=NONE;
                        d[itemNum]=0;
                    } else {
                        try {
                            charNum=nn[1];
                            thisChar=actionLine.charAt(charNum);
                        } catch (ArrayIndexOutOfBoundsException aioobe) {
// System.out.println("got to ... looks like a nonaction 2");
                            a[itemNum]=ROBOT_ACTION;
                            b[itemNum]=currentRobotId;
                            c[itemNum]=NONE;
                            d[itemNum]=0;
                        }
                    }
                } else {
                    charNum=nn[1];
                    thisChar=actionLine.charAt(charNum+1);
                    if (thisChar=='#') {
// System.out.println("got to ... looks like a nonaction 3");
                        a[itemNum]=ROBOT_ACTION;
                        b[itemNum]=currentRobotId;
                        c[itemNum]=NONE;
                        d[itemNum]=0;
                    } else {
// System.out.println("got to ... looks like fun action");
                     thisChar=actionLine.charAt(charNum);
//                     System.out.println("THIS CHAR IS " + thisChar);
                    }
                }
            }
//            System.out.println("halfway through, thisChar='" + thisChar + "'");
            if (thisChar=='N') {
// System.out.println("got to N");
                a[itemNum]=ROBOT_ACTION;
                b[itemNum]=currentRobotId;
                c[itemNum]=MOVE;
                d[itemNum]=NORTH;
                charNum++;
            }
            if (thisChar=='S') {
// System.out.println("got to S");
                a[itemNum]=ROBOT_ACTION;
                b[itemNum]=currentRobotId;
                c[itemNum]=MOVE;
                d[itemNum]=SOUTH;
                charNum++;
            }
            if (thisChar=='E') {
// System.out.println("got to E");
                a[itemNum]=ROBOT_ACTION;
                b[itemNum]=currentRobotId;
                c[itemNum]=MOVE;
                d[itemNum]=EAST;
                charNum++;
            }
            if (thisChar=='W') {
// System.out.println("got to W");
                a[itemNum]=ROBOT_ACTION;
                b[itemNum]=currentRobotId;
                c[itemNum]=MOVE;
                d[itemNum]=WEST;
                charNum++;
            }
            if (thisChar=='P') {
// System.out.println("got to P");
                int[] nn=scanTillNotInt(actionLine, charNum+2);
                a[itemNum]=ROBOT_ACTION; 
                b[itemNum]=currentRobotId;
                c[itemNum]=PICK;
                d[itemNum]=nn[0];
                charNum=nn[1];
            }
            if (thisChar=='D') {
// System.out.println("got to D");
                int[] nn=scanTillNotInt(actionLine, charNum+2);
                a[itemNum]=ROBOT_ACTION; 
                b[itemNum]=currentRobotId;
                c[itemNum]=DROP;
                d[itemNum]=nn[0];
                charNum=nn[1];
            }
            if (thisChar=='X') {
// System.out.println("got to X");
                int[] nn=scanTillNotInt(actionLine, charNum+2);
                a[itemNum]=NEW_ROBOT_POSITION; 
                b[itemNum]=currentRobotId;
                c[itemNum]=nn[0];
                nn=scanTillNotInt(actionLine, nn[1]+2);
                d[itemNum]=nn[0];
                charNum=nn[1];
            }
        }
        
        // now the package line
        System.out.println("Parsing packageLine '" + packageLine + "'");
        
        StringTokenizer st=new StringTokenizer(packageLine, " ");
        int xyz=0;
        try {
            while (st.hasMoreTokens()) {
                a[numThings+xyz]=Integer.parseInt(st.nextToken());
                b[numThings+xyz]=Integer.parseInt(st.nextToken());
                c[numThings+xyz]=Integer.parseInt(st.nextToken());
                d[numThings+xyz]=Integer.parseInt(st.nextToken());
                xyz++;
            }
        } catch (NumberFormatException nfe) {
            System.out.println("UH OH, bad bad bad.");
        }
//        System.out.println("HIIIIIIIIIIIIIIIIIIII");
        
        for (int zz=0; zz<numThings+numPackages; zz++) {
            System.out.println(a[zz] + " " + b[zz] + " " + c[zz] + " " + d[zz] );
        }
        
        robot.respond(a,b,c,d); 
/*        

#2 E #3 #1 N Robot 2 moved east, robot 1 moved north, and robot 3 did not move at all.  
#1 W #2 P 0 P 1 Robot 2 picked up package 0 and 1 and robot 1 moved west.  
#2 E #1 W E Robot 2 moved east and robot 1 moved west and then east (i.e. robot 1 was pushed).  
#1 X 1 Y 1 E #2 E Robot 2 moved east and robot 1 appeared at location (1,1) and moved east.  


N The robot moved north. 
S The robot moved south. 
E The robot moved east. 
W The robot moved west. 
P id The robot picked up package with id id. 
D id The robot dropped package with id id. 
X num1 Y num2 The robots location is set to (num1,num2). 


2 packages (id, xDest, yDest, weight)

17 133 28 50 89 11 57 78



    // a                   b         c          d
    // ---------------------------------------------------------------
    // NEW_ROBOT_POSITION  robotId   xPosition  yPosition
    // ROBOT_ACTION        robotId   PICK       packageId
    // ROBOT_ACTION        robotId   DROP       packageId
    // ROBOT_ACTION        robotId   MOVE       NORTH|SOUTH|EAST|WEST
    // ROBOT_ACTION        robotId   NONE
    // packageId           weight    destX      destY
    //
    // Note: 
    //        - a is a packageId if it's non-negative.
    //        - constants are from Simulator
    public void respond(int[] a, int[] b, int[] c, int[] d) {
        // 
    }
        
 */       
       
    }
   
    // {theInt, charNumAfterInt}
    public int[] scanTillNotInt(String s, int cn) {
        int charNum=cn;
        StringBuffer b=new StringBuffer();
        boolean lastWasNumeric=true;
        while ((lastWasNumeric)&&(charNum<s.length())) {
            char c=s.charAt(charNum++);
            System.out.println(c);
            if (c=='0'||c=='1'||c=='2'||c=='3'||c=='4'||c=='5'||c=='6'||c=='7'||c=='8'||c=='9') {
                b.append(c);
                lastWasNumeric=true;
            } else {
                lastWasNumeric=false;
            }
        }
        int num=0;
        try {
            num=Integer.parseInt(b.toString());
        } catch (NumberFormatException nfe) {
            System.out.println("That wasn't supposed to happen: '" + b.toString() + "'");
        }
        return new int[] {num, charNum};
    }
    
    
    
    public int numOccurs(char c, String s) {
        int r=0;
        for (int i=0; i<s.length(); i++) {
            if (s.charAt(i)==c) r++;
        }
        return r;
    }

    // #n+[end| #]
    public int numNonActions(String s) {
        int r=0;
        StringTokenizer st = new StringTokenizer(s, "#", false);
        while (st.hasMoreTokens()) {
            String n=st.nextToken();
            if (numOccurs(' ', n)<2) {
                char c=n.charAt(n.length()-1);
                if ((c!='N') && (c!='S') && (c!='E') && (c!='W')) {
                    r++;
                }
            }
        }
        return r;
    }
    
    public Iterator players() {
        return m_players.iterator();
    }
    
    public void run() {
        // for each robot, send it's queued move to it's associated
        // connection.  (if the connection was closed, print the score
        // for that robot and remove it from the list of players)
        // then, tell each player to decide what to do.
        while (m_players.size()>0) {
            Iterator iter=m_players.iterator();
            while (iter.hasNext()) {
                AwareRobot r=(AwareRobot) iter.next();
        System.out.println(r.getStats());
                DataOutputStream o=(DataOutputStream) m_dataOutputStreams.get(r);
                BufferedReader i=(BufferedReader) m_bufferedReaders.get(r);
                String sendText=(String) m_nextRequestedAction.get(r);
                try {
                    o.writeBytes(sendText + "\n");
                    sendServerReply(i, r);
                } catch (IOException ex) {
                    System.out.println("IOException during Simulator.run(): " + ex.getMessage());
                    try {o.close();
                         i.close(); } catch (Exception ex2) { }
                    m_players.remove(r);
                }
            }
        }
    }
    
    public static void main(String[] args) throws Exception {
        ICFPSimulatorProxy sim=new ICFPSimulatorProxy(args[0], Integer.parseInt(args[1]));
        AwareRobot robot=sim.addPlayer(new Genome());
        sim.run();
    }
    
}